/*! \file scalarfield.cpp
 *  \brief Source code for scalarfield.cpp
 */

/* Copyright (c) 2005-2009 Taneli Kalvas. All rights reserved.
 *
 * You can redistribute this software and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this library (file "COPYING" included in the package);
 * if not, write to the Free Software Foundation, Inc., 51 Franklin
 * Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * If you have questions about your rights to use or distribute this
 * software, please contact Berkeley Lab's Technology Transfer
 * Department at TTD@lbl.gov. Other questions, comments and bug
 * reports should be sent directly to the author via email at
 * taneli.kalvas@jyu.fi.
 * 
 * NOTICE. This software was developed under partial funding from the
 * U.S.  Department of Energy.  As such, the U.S. Government has been
 * granted for itself and others acting on its behalf a paid-up,
 * nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and
 * display publicly.  Beginning five (5) years after the date
 * permission to assert copyright is obtained from the U.S. Department
 * of Energy, and subject to any subsequent five (5) year renewals,
 * the U.S. Government is granted for itself and others acting on its
 * behalf a paid-up, nonexclusive, irrevocable, worldwide license in
 * the Software to reproduce, prepare derivative works, distribute
 * copies to the public, perform publicly and display publicly, and to
 * permit others to do so.
 */

#include "scalarfield.hpp"
#include <iostream>
#include <string.h>
#include <cmath>
#include <limits>


ScalarField::ScalarField( const Geometry &g )
    : _geom_mode(g.geom_mode()), _size(g.size()), _origo(g.origo()), _h(g.h())
{
    // Calculate vector max
    _max = Vec3D( _origo(0)+_h*(_size[0]-1),
		  _origo(1)+_h*(_size[1]-1),
		  _origo(2)+_h*(_size[2]-1) );
    _div_h = 1.0/_h;
    _F = new double[_size[0]*_size[1]*_size[2]];
    memset( _F, 0, _size[0]*_size[1]*_size[2]*sizeof(double) );
}


ScalarField::ScalarField( geom_mode_e geom_mode, Int3D size, Vec3D origo, double h )
    : _geom_mode(geom_mode), _size(size), _origo(origo), _h(h)
{
    if( _geom_mode == MODE_3D ) {
	if( _size[0] < 3 || _size[1] < 3 || _size[2] < 3 )
	    throw( Error( ERROR_LOCATION, "illegal mesh size" ) );
    } else if( _geom_mode == MODE_2D || _geom_mode == MODE_CYL ) {
	if( _size[0] < 3 || _size[1] < 3 || _size[2] != 1 )
	    throw( Error( ERROR_LOCATION, "illegal mesh size" ) );
    } else {
	if( _size[0] < 3 || _size[1] != 1 || _size[2] != 1 )
	    throw( Error( ERROR_LOCATION, "illegal mesh size" ) );
    }

    // Calculate vector max
    _max = Vec3D( _origo(0)+_h*(_size[0]-1),
		  _origo(1)+_h*(_size[1]-1),
		  _origo(2)+_h*(_size[2]-1) );
    _div_h = 1.0/_h;
    _F = new double[_size[0]*_size[1]*_size[2]];
    memset( _F, 0, _size[0]*_size[1]*_size[2]*sizeof(double) );
}


ScalarField::ScalarField( std::istream &s )
{
    _geom_mode = (geom_mode_e)read_int32( s );
    _size      = Int3D( s );
    _origo     = Vec3D( s );
    _h         = read_double( s );
    // Calculate vector max
    _max = Vec3D( _origo(0)+_h*(_size[0]-1),
		  _origo(1)+_h*(_size[1]-1),
		  _origo(2)+_h*(_size[2]-1) );
    _div_h     = 1.0/_h;
    _F = new double[_size[0]*_size[1]*_size[2]];
    read_compressed_block( s, _size[0]*_size[1]*_size[2]*sizeof(double), (int8_t *)_F );
}


ScalarField::ScalarField( const ScalarField &f )
    : _geom_mode(f._geom_mode), _size(f._size), _origo(f._origo), _max(f._max), _h(f._h), 
      _div_h(f._div_h)
{
    _F = new double[_size[0]*_size[1]*_size[2]];
    memcpy( _F, f._F, _size[0]*_size[1]*_size[2]*sizeof(double) );
}


ScalarField::~ScalarField()
{
    delete [] _F;
}


void ScalarField::clear()
{
    memset( _F, 0, _size[0]*_size[1]*_size[2]*sizeof(double) );
}


void ScalarField::reset( geom_mode_e geom_mode, Int3D size, 
			 Vec3D origo, double h )
{
    _geom_mode = geom_mode;
    _size      = size;
    _origo     = origo;
    _h         = h;
    _max = Vec3D( _origo(0)+_h*(_size[0]-1),
		  _origo(1)+_h*(_size[1]-1),
		  _origo(2)+_h*(_size[2]-1) );
    _div_h     = 1.0/h;
    delete [] _F;
    _F = new double[_size[0]*_size[1]*_size[2]];
    memset( _F, 0, _size[0]*_size[1]*_size[2]*sizeof(double) );
}


void ScalarField::get_minmax( double &min, double &max ) const
{
    min = _F[0];
    max = _F[0];

    size_t ncount = _size[0]*_size[1]*_size[2];
    std::cout << "ncount = " << ncount << "\n";
    for( size_t a = 1; a < ncount; a++ ) {
	if( _F[a] < min )
	    min = _F[a];
	if( _F[a] > max )
	    max = _F[a];
	std::cout << "a = " << a << ": " 
		  << "min = " << min << " "
		  << "max = " << max << "\n";
    }
}


void ScalarField::epot_get_minmax( const Geometry &g, double &min, double &max ) const
{
    min = std::numeric_limits<double>::infinity();
    max = -std::numeric_limits<double>::infinity();

    size_t ncount = _size[0]*_size[1]*_size[2];
    for( size_t a = 0; a < ncount; a++ ) {
	if( g.mesh( a ) <= -7 )
	    // Skip electrode edges
	    continue; 
	if( _F[a] < min )
	    min = _F[a];
	if( _F[a] > max )
	    max = _F[a];
    }
}


ScalarField &ScalarField::operator=( const ScalarField &f )
{
    _geom_mode = f._geom_mode;
    _size      = f._size;
    _origo     = f._origo;
    _h         = f._h;
    _div_h     = f._div_h;
    delete [] _F;
    _F = new double[_size[0]*_size[1]*_size[2]];
    memcpy( _F, f._F, _size[0]*_size[1]*_size[2]*sizeof(double) );
    return( *this );
}


ScalarField &ScalarField::operator+=( const ScalarField &f )
{
    if( _geom_mode != f._geom_mode || _size != f._size ||
	_origo != f._origo || _h != f._h )
	throw( Error( ERROR_LOCATION, "non-matching fields" ) );
    size_t ncount = _size[0]*_size[1]*_size[2];
    for( size_t a = 0; a < ncount; a++ )
	_F[a] += f._F[a];
    return( *this );
}


ScalarField &ScalarField::operator*=( double x )
{
    size_t ncount = _size[0]*_size[1]*_size[2];
    for( size_t a = 0; a < ncount; a++ )
	_F[a] *= x;
    return( *this );
}


ScalarField &ScalarField::operator/=( double x )
{
    double xi = 1.0/x;
    size_t ncount = _size[0]*_size[1]*_size[2];
    for( size_t a = 0; a < ncount; a++ )
	_F[a] *= xi;
    return( *this );
}


double ScalarField::operator()( Vec3D x ) const
{
    switch( _geom_mode ) {
    case MODE_1D:
    {
	int32_t i = (int32_t)floor( (x[0]-_origo[0])*_div_h );
	if( i < 0 )
	    i = 0;
	else if( i >= _size[0]-1 )
	    i = _size[0]-2;

	double t = _div_h*( x[0]-(i*_h+_origo[0]) );

	return( (1.0-t)*_F[i] + t*_F[i+1] );
	break;
    }
    case MODE_2D:
    case MODE_CYL:
    {
	int32_t i = (int32_t)floor( (x[0]-_origo[0])*_div_h );
	int32_t j = (int32_t)floor( (x[1]-_origo[1])*_div_h );
	if( i < 0 )
	    i = 0;
	else if( i >= _size[0]-1 )
	    i = _size[0]-2;
	if( j < 0 )
	    j = 0;
	else if( j >= _size[1]-1 )
	    j = _size[1]-2;

	double t = _div_h*( x[0]-(i*_h+_origo[0]) );
	double u = _div_h*( x[1]-(j*_h+_origo[1]) );

	int32_t ptr = _size[0]*j + i;
	return( (1.0-t)*(1.0-u)*_F[ptr] +
		(    t)*(1.0-u)*_F[ptr+1] +
		(1.0-t)*(    u)*_F[ptr+_size[0]] +
		(    t)*(    u)*_F[ptr+1+_size[0]] );
	break;
    }
    default:
    {
	int32_t i = (int32_t)floor( (x[0]-_origo[0])*_div_h );
	int32_t j = (int32_t)floor( (x[1]-_origo[1])*_div_h );
	int32_t k = (int32_t)floor( (x[2]-_origo[2])*_div_h );
	if( i < 0 )
	    i = 0;
	else if( i >= _size[0]-1 )
	    i = _size[0]-2;
	if( j < 0 )
	    j = 0;
	else if( j >= _size[1]-1 )
	    j = _size[1]-2;
	if( k < 0 )
	    k = 0;
	else if( k >= _size[2]-1 )
	    k = _size[2]-2;

	double t = _div_h*( x[0]-(i*_h+_origo[0]) );
	double u = _div_h*( x[1]-(j*_h+_origo[1]) );
	double v = _div_h*( x[2]-(k*_h+_origo[2]) );

	int32_t b  = _size[0]*_size[1];
	int32_t ptr = b*k + _size[0]*j + i;
	return( (1.0-t)*(1.0-u)*(1.0-v)*_F[ptr] +
		(    t)*(1.0-u)*(1.0-v)*_F[ptr+1] +
		(1.0-t)*(    u)*(1.0-v)*_F[ptr+_size[0]] +
		(    t)*(    u)*(1.0-v)*_F[ptr+1+_size[0]] +
		(1.0-t)*(1.0-u)*(    v)*_F[ptr+b] +
		(    t)*(1.0-u)*(    v)*_F[ptr+1+b] +
		(1.0-t)*(    u)*(    v)*_F[ptr+_size[0]+b] +
		(    t)*(    u)*(    v)*_F[ptr+1+_size[0]+b] );
	break;
    }
    }

}


void ScalarField::save( std::ostream &s ) const
{
    write_int32( s, _geom_mode );
    _size.save( s );
    _origo.save( s );
    write_double( s, _h );
    write_compressed_block( s, _size[0]*_size[1]*_size[2]*sizeof(double), (int8_t *)_F );
}


void ScalarField::debug_print( void ) const
{
    int32_t a;
    std::cout << "**ScalarField\n";
    if( _geom_mode == MODE_1D )
	std::cout << "geom_mode = MODE_1D\n";
    else if( _geom_mode == MODE_2D )
	std::cout << "geom_mode = MODE_2D\n";
    else if( _geom_mode == MODE_CYL )
	std::cout << "geom_mode = MODE_CYL\n";
    else if( _geom_mode == MODE_3D )
	std::cout << "geom_mode = MODE_3D\n";
    else
	std::cout << "geom_mode = Unknown\n";
    std::cout << "size = (" 
	      << _size[0] << ", "
	      << _size[1] << ", "
	      << _size[2] << ")\n";
    std::cout << "origo = (" 
	      << _origo[0] << ", "
	      << _origo[1] << ", "
	      << _origo[2] << ")\n";
    std::cout << "h = " << _h << "\n";
    std::cout << "div_h = " << _div_h << "\n";
    std::cout << "F = (";
    for( a = 0; a < _size[0]*_size[1]*_size[2]-1; a++ )
	std::cout << _F[a] << ", ";
    if( a < _size[0]*_size[1]*_size[2] )
	std::cout << _F[a] << ")\n";

}




















